//
// Copyright (c) 2004-2006, Skype Limited.
// All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
//   * Redistributions of source code must retain the above copyright
//     notice, this list of conditions and the following disclaimer.
//   * Redistributions in binary form must reproduce the above
//     copyright notice, this list of conditions and the following
//     disclaimer in the documentation and/or other materials provided
//     with the distribution.
//   * Neither the name of the Skype Limited nor the names of its
//     contributors may be used to endorse or promote products derived
//     from this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
// COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
// ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//

// This source code is based on msgapitest.cpp example from Skype SDK

#include <process.h>
#include <windows.h>
#include <rpcdce.h>
#include "main.h"

MSG_CALLBACK mcb;
HANDLE Thread;
HANDLE Mutex;


HWND hInit_MainWindowHandle;
HINSTANCE hInit_ProcessHandle;
char acInit_WindowClassName[128];
HANDLE hGlobal_ThreadShutdownEvent;
bool volatile fGlobal_ThreadRunning=true;
UINT uiGlobal_MsgID_SkypeControlAPIAttach;
UINT uiGlobal_MsgID_SkypeControlAPIDiscover;
HWND hGlobal_SkypeAPIWindowHandle=NULL;

bool volatile bLoaded = false;
char cmd[1024] = {0};

enum
{
    SKYPECONTROLAPI_ATTACH_SUCCESS=0,								// Client is successfully attached and API window handle can be found in wParam parameter
    SKYPECONTROLAPI_ATTACH_PENDING_AUTHORIZATION=1,	// Skype has acknowledged connection request and is waiting for confirmation from the user.
    // The client is not yet attached and should wait for SKYPECONTROLAPI_ATTACH_SUCCESS message
    SKYPECONTROLAPI_ATTACH_REFUSED=2,								// User has explicitly denied access to client
    SKYPECONTROLAPI_ATTACH_NOT_AVAILABLE=3,					// API is not available at the moment. For example, this happens when no user is currently logged in.
    // Client should wait for SKYPECONTROLAPI_ATTACH_API_AVAILABLE broadcast before making any further
    // connection attempts.
    SKYPECONTROLAPI_ATTACH_API_AVAILABLE=0x8001
};

static LRESULT APIENTRY SkypeAPITest_Windows_WindowProc(
    HWND hWindow, UINT uiMessage, WPARAM uiParam, LPARAM ulParam)
{
    LRESULT lReturnCode;
    bool fIssueDefProc;

    lReturnCode=0;
    fIssueDefProc=false;
    switch (uiMessage)
    {
    case WM_DESTROY:
        hInit_MainWindowHandle=NULL;
        PostQuitMessage(0);
        break;
    case WM_COPYDATA:
        if ( hGlobal_SkypeAPIWindowHandle==(HWND)uiParam )
        {
            //PCOPYDATASTRUCT poCopyData=(PCOPYDATASTRUCT)ulParam;
            //printf( "Message from Skype(%u): %.*s\n", poCopyData->dwData, poCopyData->cbData, poCopyData->lpData);
            lReturnCode=1;
        }
        break;
    default:
        if ( uiMessage==uiGlobal_MsgID_SkypeControlAPIAttach )
        {
            switch (ulParam)
            {
            case SKYPECONTROLAPI_ATTACH_SUCCESS:
                OutputDebugStr("Skype: Connected\n");
                hGlobal_SkypeAPIWindowHandle=(HWND)uiParam;
                break;
            case SKYPECONTROLAPI_ATTACH_PENDING_AUTHORIZATION:
                OutputDebugStr("Skype: Pending authorization\n");
                break;
            case SKYPECONTROLAPI_ATTACH_REFUSED:
                OutputDebugStr("Skype: Connection refused\n");
                break;
            case SKYPECONTROLAPI_ATTACH_NOT_AVAILABLE:
                OutputDebugStr("Skype: API not available\n");
                break;
            case SKYPECONTROLAPI_ATTACH_API_AVAILABLE:
                OutputDebugStr("Skype: Try connect now (API available);\n");
                break;
            }
            lReturnCode=1;
            break;
        }
        fIssueDefProc=true;
        break;
    }
    if ( fIssueDefProc )
        lReturnCode=DefWindowProc( hWindow, uiMessage, uiParam, ulParam);
    return(lReturnCode);
}

bool Initialize_CreateWindowClass(void)
{
    unsigned char *paucUUIDString;
    RPC_STATUS lUUIDResult;
    bool fReturnStatus;
    UUID oUUID;

    fReturnStatus=false;
    lUUIDResult=UuidCreate(&oUUID);
    hInit_ProcessHandle=(HINSTANCE)OpenProcess( PROCESS_DUP_HANDLE, FALSE, GetCurrentProcessId());
    if ( hInit_ProcessHandle!=NULL && (lUUIDResult==RPC_S_OK || lUUIDResult==RPC_S_UUID_LOCAL_ONLY) )
    {
        if ( UuidToString( &oUUID, &paucUUIDString)==RPC_S_OK )
        {
            WNDCLASS oWindowClass;

            strcpy( acInit_WindowClassName, "FreeGgStatusToSkype");
            strcat( acInit_WindowClassName, (char *)paucUUIDString);

            oWindowClass.style=CS_HREDRAW|CS_VREDRAW|CS_DBLCLKS;
            oWindowClass.lpfnWndProc=(WNDPROC)&SkypeAPITest_Windows_WindowProc;
            oWindowClass.cbClsExtra=0;
            oWindowClass.cbWndExtra=0;
            oWindowClass.hInstance=hInit_ProcessHandle;
            oWindowClass.hIcon=NULL;
            oWindowClass.hCursor=NULL;
            oWindowClass.hbrBackground=NULL;
            oWindowClass.lpszMenuName=NULL;
            oWindowClass.lpszClassName=acInit_WindowClassName;

            if ( RegisterClass(&oWindowClass)!=0 )
                fReturnStatus=true;

            RpcStringFree(&paucUUIDString);
        }
    }
    if ( fReturnStatus==false )
        CloseHandle(hInit_ProcessHandle),hInit_ProcessHandle=NULL;
    return(fReturnStatus);
}

void DeInitialize_DestroyWindowClass(void)
{
    UnregisterClass( acInit_WindowClassName, hInit_ProcessHandle);
    CloseHandle(hInit_ProcessHandle),hInit_ProcessHandle=NULL;
}

bool Initialize_CreateMainWindow(void)
{
    hInit_MainWindowHandle=CreateWindowEx( WS_EX_APPWINDOW|WS_EX_WINDOWEDGE,
                                           acInit_WindowClassName, "", WS_BORDER|WS_SYSMENU|WS_MINIMIZEBOX,
                                           CW_USEDEFAULT, CW_USEDEFAULT, 128, 128, NULL, 0, hInit_ProcessHandle, 0);
    return(hInit_MainWindowHandle!=NULL? true:false);
}

void DeInitialize_DestroyMainWindow(void)
{
    if ( hInit_MainWindowHandle!=NULL )
        DestroyWindow(hInit_MainWindowHandle),hInit_MainWindowHandle=NULL;
}

void Global_MessageLoop(void)
{
    MSG oMessage;

    while (GetMessage( &oMessage, 0, 0, 0)!=FALSE)
    {
        TranslateMessage(&oMessage);
        DispatchMessage(&oMessage);
    }
}

void __cdecl Global_InputProcessingThread(void *)
{
    int iTimer = -1;
    while (bLoaded)
    {
        if ( hGlobal_SkypeAPIWindowHandle!=NULL )
        {
            if ( !((++iTimer)%30) )
            {
                COPYDATASTRUCT oCopyData;

                // send command to skype
                oCopyData.dwData=0;
                if (WaitForSingleObject(Mutex, 5000L) != WAIT_OBJECT_0) continue;
                oCopyData.lpData=cmd;
                oCopyData.cbData=strlen(cmd)+1;
                if ( oCopyData.cbData!=1 )
                {
                    if ( SendMessage( hGlobal_SkypeAPIWindowHandle, WM_COPYDATA, (WPARAM)hInit_MainWindowHandle, (LPARAM)&oCopyData)==FALSE )
                    {
                        hGlobal_SkypeAPIWindowHandle=NULL;
                        OutputDebugStr("Skype: Disconnected\n");
                    }
                }
                cmd[0] = 0;
                ReleaseMutex(Mutex);
            }
        }
        else
        {
            if ( !((++iTimer)%300) )
                SendMessage( HWND_BROADCAST, uiGlobal_MsgID_SkypeControlAPIDiscover, (WPARAM)hInit_MainWindowHandle, 0);
        }
        Sleep(40);
    }
    hGlobal_SkypeAPIWindowHandle=NULL;
    PostMessage( hInit_MainWindowHandle, WM_CLOSE, 0, 0);
    SetEvent(hGlobal_ThreadShutdownEvent);
    fGlobal_ThreadRunning=false;
}


DWORD WINAPI ThreadProc(LPVOID data)
{
    Mutex = CreateMutex(NULL,FALSE,NULL);

    // create window class
    //   create dummy/hidden window for processing messages
    //     run message loop thread
    //       do application control until exit
    //       exit: send QUIT message to our own window
    //             wait until thred terminates
    //   destroy main window
    // destroy window class
    uiGlobal_MsgID_SkypeControlAPIAttach=RegisterWindowMessage("SkypeControlAPIAttach");
    uiGlobal_MsgID_SkypeControlAPIDiscover=RegisterWindowMessage("SkypeControlAPIDiscover");
    if ( uiGlobal_MsgID_SkypeControlAPIAttach!=0 && uiGlobal_MsgID_SkypeControlAPIDiscover!=0 )
    {
        if ( Initialize_CreateWindowClass() )
        {
            if ( Initialize_CreateMainWindow() )
            {
                hGlobal_ThreadShutdownEvent=CreateEvent( NULL, TRUE, FALSE, NULL);
                if ( hGlobal_ThreadShutdownEvent!=NULL )
                {
                    if ( _beginthread( &Global_InputProcessingThread, 64*1024, NULL)!=(unsigned long)-1 )
                    {
                        Global_MessageLoop();
                        WaitForSingleObject( hGlobal_ThreadShutdownEvent, INFINITE);
                    }
                    CloseHandle(hGlobal_ThreadShutdownEvent);
                }
                DeInitialize_DestroyMainWindow();
            }
            DeInitialize_DestroyWindowClass();
        }
    }
    return 0;
}

char* Name()
{
    return "Status To Skype";
}

int Load(MSG_CALLBACK callback)
{
    DWORD dwtid;
    mcb = callback;
    bLoaded = true;
    Thread = CreateThread(NULL,0,ThreadProc,0,0,&dwtid);
    return 0;
}

int Unload()
{
    bLoaded = false;
    Sleep(50);
    TerminateThread(Thread,0);
    return 0;
}

inline bool StartsWith(char* &a, char *b)
{
    if (strlen(a) < strlen(b))
    {
        return false;
    }
    else
    {
        bool result = !strnicmp(a,b,strlen(b));
        if (result)
            a += strlen(b);
        return result;
    }
}

int HandleMessage(char* message)
{
    if (StartsWith(message,"status "))
    {
        if (StartsWith(message, "desc "))
        {
            if (WaitForSingleObject(Mutex, 5000L) != WAIT_OBJECT_0) return -1;
            strcpy(cmd, "SET PROFILE MOOD_TEXT ");
            strcat(cmd, message);
            ReleaseMutex(Mutex);
        }
    }
    return 0;
}
